@@ -8,11 +8,10 @@ pub use future_factory::*;
 
 use crate::{
     common::*,
-    config::{NumWorkers, ParParams},
+    config::ParParams,
     index_stream::{IndexStreamExt as _, ReorderEnumerated},
-    rt,
-    stream::StreamExt as _,
-    utils,
+    par_stream::ParStreamExt as _,
+    rt, utils,
 };
 use flume::r#async::RecvStream;
 use tokio::sync::broadcast;
@@ -187,7 +186,7 @@ where
             buf_size,
         } = params.into();
 
-        let stream = stream.map(move |item| fac.generate(item)).shared();
+        let stream = stream.map(move |item| fac.generate(item)).spawned(buf_size);
         let (output_tx, output_rx) = utils::channel(buf_size);
 
         (0..num_workers).for_each(move |_| {
@@ -223,7 +222,7 @@ where
         let stream = stream
             .map(move |item| fac.generate(item))
             .enumerate()
-            .shared();
+            .spawned(buf_size);
         let (output_tx, output_rx) = utils::channel(buf_size);
 
         (0..num_workers).for_each(move |_| {
@@ -251,15 +250,18 @@ where
     Fac::Fut: 'static + Send + Future<Output = ()>,
 {
     /// Runs parallel tasks on each stream item.
-    pub async fn for_each<N>(self, num_workers: N)
+    pub async fn for_each<P>(self, params: P)
     where
-        N: Into<NumWorkers>,
+        P: Into<ParParams>,
     {
-        let num_workers = num_workers.into().get();
+        let ParParams {
+            num_workers,
+            buf_size,
+        } = params.into();
         let Self {
             mut fac, stream, ..
         } = self;
-        let stream = stream.map(move |item| fac.generate(item)).shared();
+        let stream = stream.map(move |item| fac.generate(item)).spawned(buf_size);
 
         let worker_futures = (0..num_workers).map(move |_| {
             let stream = stream.clone();
@@ -282,11 +284,14 @@ where
     Error: 'static + Send,
 {
     /// Runs parallel tasks on each stream item.
-    pub async fn try_for_each<N>(self, num_workers: N) -> Result<(), Error>
+    pub async fn try_for_each<P>(self, params: P) -> Result<(), Error>
     where
-        N: Into<NumWorkers>,
+        P: Into<ParParams>,
     {
-        let num_workers = num_workers.into().get();
+        let ParParams {
+            num_workers,
+            buf_size,
+        } = params.into();
         let Self {
             mut fac, stream, ..
         } = self;
@@ -296,7 +301,7 @@ where
                 let _ = terminate_rx.recv().await;
             })
             .map(move |item| fac.generate(item))
-            .shared();
+            .spawned(buf_size);
 
         let worker_futures = (0..num_workers).map(move |_| {
             let stream = stream.clone();
@@ -390,7 +395,7 @@ where
             buf_size,
         } = params.into();
 
-        let stream = stream.map(move |item| fac.generate(item)).shared();
+        let stream = stream.map(move |item| fac.generate(item)).spawned(buf_size);
         let (output_tx, output_rx) = utils::channel(buf_size);
 
         (0..num_workers).for_each(move |_| {
@@ -428,7 +433,7 @@ where
         let stream = stream
             .map(move |item| fac.generate(item))
             .enumerate()
-            .shared();
+            .spawned(buf_size);
         let (output_tx, output_rx) = utils::channel(buf_size);
 
         (0..num_workers).for_each(move |_| {
@@ -455,18 +460,21 @@ where
     St: 'static + Send + Stream,
     St::Item: 'static + Send,
     Fac: 'static + Send + FnFactory<St::Item, ()>,
-    Fac::Fn: 'static + Send + FnOnce() -> (),
+    Fac::Fn: 'static + Send + FnOnce(),
 {
     /// Runs parallel tasks on each stream item.
-    pub async fn for_each<N>(self, num_workers: N)
+    pub async fn for_each<P>(self, params: P)
     where
-        N: Into<NumWorkers>,
+        P: Into<ParParams>,
     {
         let Self {
             mut fac, stream, ..
         } = self;
-        let num_workers = num_workers.into().get();
-        let stream = stream.map(move |item| fac.generate(item)).shared();
+        let ParParams {
+            num_workers,
+            buf_size,
+        } = params.into();
+        let stream = stream.map(move |item| fac.generate(item)).spawned(buf_size);
 
         let worker_futures = (0..num_workers).map(move |_| {
             let mut stream = stream.clone();
@@ -640,14 +648,14 @@ where
     FutFac::Fut: 'static + Send + Future,
     <FutFac::Fut as Future>::Output: 'static + Send,
     FnFac: 'static + Send + Clone + FnFactory<<FutFac::Fut as Future>::Output, ()>,
-    FnFac::Fn: 'static + Send + FnOnce() -> (),
+    FnFac::Fn: 'static + Send + FnOnce(),
 {
     /// Runs parallel tasks on each stream item.
-    pub async fn for_each<N>(self, num_workers: N)
+    pub async fn for_each<P>(self, params: P)
     where
-        N: Into<NumWorkers>,
+        P: Into<ParParams>,
     {
-        self.into_async_builder().for_each(num_workers).await;
+        self.into_async_builder().for_each(params).await;
     }
 }
 
@@ -673,42 +681,13 @@ where
 #[cfg(test)]
 mod tests {
     use super::*;
-    use crate::par_stream::ParStreamExt as _;
+    use crate::utils::async_test;
 
-    #[tokio::test]
-    async fn par_builder_blocking_test() {
-        let vec: Vec<_> = stream::iter(1u64..=1000)
-            .par_builder()
-            .map_blocking(|val| move || val.pow(5))
-            .map_blocking(|val| move || val + 1)
-            .build_ordered_stream(None)
-            .collect()
-            .await;
-        let expect: Vec<_> = (1u64..=1000).map(|val| val.pow(5) + 1).collect();
-
-        assert_eq!(vec, expect);
-    }
-
-    #[tokio::test]
-    async fn par_builder_async_test() {
-        let vec: Vec<_> = stream::iter(1u64..=1000)
-            .par_builder()
-            .map_async(|val| async move { val.pow(5) })
-            .map_async(|val| async move { val + 1 })
-            .build_ordered_stream(None)
-            .collect()
-            .await;
-        let expect: Vec<_> = (1u64..=1000).map(|val| val.pow(5) + 1).collect();
-
-        assert_eq!(vec, expect);
-    }
-
-    #[tokio::test]
-    async fn par_builder_mixed_async_blocking_test() {
-        {
+    async_test! {
+        async fn par_builder_blocking_test() {
             let vec: Vec<_> = stream::iter(1u64..=1000)
                 .par_builder()
-                .map_async(|val| async move { val.pow(5) })
+                .map_blocking(|val| move || val.pow(5))
                 .map_blocking(|val| move || val + 1)
                 .build_ordered_stream(None)
                 .collect()
@@ -718,10 +697,10 @@ mod tests {
             assert_eq!(vec, expect);
         }
 
-        {
+        async fn par_builder_async_test() {
             let vec: Vec<_> = stream::iter(1u64..=1000)
                 .par_builder()
-                .map_blocking(|val| move || val.pow(5))
+                .map_async(|val| async move { val.pow(5) })
                 .map_async(|val| async move { val + 1 })
                 .build_ordered_stream(None)
                 .collect()
@@ -731,39 +710,67 @@ mod tests {
             assert_eq!(vec, expect);
         }
 
-        {
-            let vec: Vec<_> = stream::iter(1u64..=1000)
-                .par_builder()
-                .map_blocking(|val| move || val.pow(5))
-                .map_async(|val| async move { val + 1 })
-                .map_blocking(|val| move || val / 2)
-                .build_ordered_stream(None)
-                .collect()
-                .await;
-            let expect: Vec<_> = (1u64..=1000).map(|val| (val.pow(5) + 1) / 2).collect();
+        async fn par_builder_mixed_async_blocking_test() {
+            {
+                let vec: Vec<_> = stream::iter(1u64..=1000)
+                    .par_builder()
+                    .map_async(|val| async move { val.pow(5) })
+                    .map_blocking(|val| move || val + 1)
+                    .build_ordered_stream(None)
+                    .collect()
+                    .await;
+                let expect: Vec<_> = (1u64..=1000).map(|val| val.pow(5) + 1).collect();
 
-            assert_eq!(vec, expect);
+                assert_eq!(vec, expect);
+            }
+
+            {
+                let vec: Vec<_> = stream::iter(1u64..=1000)
+                    .par_builder()
+                    .map_blocking(|val| move || val.pow(5))
+                    .map_async(|val| async move { val + 1 })
+                    .build_ordered_stream(None)
+                    .collect()
+                    .await;
+                let expect: Vec<_> = (1u64..=1000).map(|val| val.pow(5) + 1).collect();
+
+                assert_eq!(vec, expect);
+            }
+
+            {
+                let vec: Vec<_> = stream::iter(1u64..=1000)
+                    .par_builder()
+                    .map_blocking(|val| move || val.pow(5))
+                    .map_async(|val| async move { val + 1 })
+                    .map_blocking(|val| move || val / 2)
+                    .build_ordered_stream(None)
+                    .collect()
+                    .await;
+                let expect: Vec<_> = (1u64..=1000).map(|val| (val.pow(5) + 1) / 2).collect();
+
+                assert_eq!(vec, expect);
+            }
+
+            {
+                let vec: Vec<_> = stream::iter(1u64..=1000)
+                    .par_builder()
+                    .map_async(|val| async move { val.pow(5) })
+                    .map_blocking(|val| move || val + 1)
+                    .map_async(|val| async move { val / 2 })
+                    .build_ordered_stream(None)
+                    .collect()
+                    .await;
+                let expect: Vec<_> = (1u64..=1000).map(|val| (val.pow(5) + 1) / 2).collect();
+
+                assert_eq!(vec, expect);
+            }
         }
 
-        {
-            let vec: Vec<_> = stream::iter(1u64..=1000)
-                .par_builder()
-                .map_async(|val| async move { val.pow(5) })
-                .map_blocking(|val| move || val + 1)
-                .map_async(|val| async move { val / 2 })
-                .build_ordered_stream(None)
-                .collect()
-                .await;
-            let expect: Vec<_> = (1u64..=1000).map(|val| (val.pow(5) + 1) / 2).collect();
+        // #[tokio::test]
+        // async fn par_unfold_builder_async_test() {
+        //     let vec: Vec<_> = super::par_unfold_builder(|| async move {
 
-            assert_eq!(vec, expect);
-        }
+        //     }).into_stream().collect();
+        // }
     }
-
-    // #[tokio::test]
-    // async fn par_unfold_builder_async_test() {
-    //     let vec: Vec<_> = super::par_unfold_builder(|| async move {
-
-    //     }).into_stream().collect();
-    // }
 }
