@@ -2,11 +2,10 @@ use crate::{
     broadcast::BroadcastBuilder,
     builder::ParBuilder,
     common::*,
-    config::{BufSize, NumWorkers, ParParams},
+    config::{BufSize, ParParams},
     index_stream::{IndexStreamExt as _, ReorderEnumerated},
     pull::PullBuilder,
     rt,
-    shared_stream::Shared,
     stream::StreamExt as _,
     tee::Tee,
     utils,
@@ -115,10 +114,10 @@ where
     /// use par_stream::prelude::*;
     ///
     /// let data = vec![1, 2, -3, 4, 5, -6, 7, 8];
-    /// stream::iter(data).par_batching(None, |_worker_index, mut stream| async move {
-    ///     while let Some(value) = stream.next().await {
+    /// stream::iter(data).par_batching(None, |_worker_index, rx| async move {
+    ///     while let Ok(value) = rx.recv_async().await {
     ///         if value > 0 {
-    ///             return Some((value, stream));
+    ///             return Some((value, rx));
     ///         }
     ///     }
     ///     None
@@ -127,8 +126,9 @@ where
     /// ```
     fn par_batching<T, P, F, Fut>(self, params: P, f: F) -> RecvStream<'static, T>
     where
-        F: 'static + Send + Clone + FnMut(usize, Shared<Self>) -> Fut,
-        Fut: 'static + Future<Output = Option<(T, Shared<Self>)>> + Send,
+        Self: Sized,
+        F: 'static + Send + Clone + FnMut(usize, flume::Receiver<Self::Item>) -> Fut,
+        Fut: 'static + Future<Output = Option<(T, flume::Receiver<Self::Item>)>> + Send,
         T: 'static + Send,
         P: Into<ParParams>;
 
@@ -346,29 +346,29 @@ where
     /// assert_eq!(sum, Some((1 + 1000) * 1000 / 2));
     /// # })
     /// ```
-    fn par_reduce<N, F, Fut>(
+    fn par_reduce<P, F, Fut>(
         self,
-        num_workers: N,
+        params: P,
         reduce_fn: F,
     ) -> BoxFuture<'static, Option<Self::Item>>
     where
-        N: Into<NumWorkers>,
+        P: Into<ParParams>,
         F: 'static + FnMut(Self::Item, Self::Item) -> Fut + Send + Clone,
         Fut: 'static + Future<Output = Self::Item> + Send;
 
     /// Runs an asynchronous task on parallel workers.
-    fn par_for_each<N, F, Fut>(self, num_workers: N, f: F) -> BoxFuture<'static, ()>
+    fn par_for_each<P, F, Fut>(self, params: P, f: F) -> BoxFuture<'static, ()>
     where
         F: 'static + FnMut(Self::Item) -> Fut + Send,
         Fut: 'static + Future<Output = ()> + Send,
-        N: Into<NumWorkers>;
+        P: Into<ParParams>;
 
     /// Runs a blocking task on parallel workers.
-    fn par_for_each_blocking<N, F, Func>(self, num_workers: N, f: F) -> BoxFuture<'static, ()>
+    fn par_for_each_blocking<P, F, Func>(self, params: P, f: F) -> BoxFuture<'static, ()>
     where
         F: 'static + FnMut(Self::Item) -> Func + Send,
         Func: 'static + FnOnce() + Send,
-        N: Into<NumWorkers>;
+        P: Into<ParParams>;
 }
 
 impl<S> ParStreamExt for S
@@ -399,15 +399,12 @@ where
         let mut stream = self.boxed();
         let (output_tx, output_rx) = utils::channel(buf_size);
 
-        rt::spawn_blocking(move || loop {
-            match rt::block_on(stream.next()) {
-                Some(input) => {
-                    let output = f(input);
-                    if output_tx.send(output).is_err() {
-                        break;
-                    }
+        rt::spawn_blocking(move || {
+            while let Some(input) = rt::block_on(stream.next()) {
+                let output = f(input);
+                if output_tx.send(output).is_err() {
+                    break;
                 }
-                None => break,
             }
         });
 
@@ -420,8 +417,8 @@ where
 
     fn par_batching<T, P, F, Fut>(self, params: P, f: F) -> RecvStream<'static, T>
     where
-        F: 'static + Send + Clone + FnMut(usize, Shared<Self>) -> Fut,
-        Fut: 'static + Future<Output = Option<(T, Shared<Self>)>> + Send,
+        F: 'static + Send + Clone + FnMut(usize, flume::Receiver<Self::Item>) -> Fut,
+        Fut: 'static + Future<Output = Option<(T, flume::Receiver<Self::Item>)>> + Send,
         T: 'static + Send,
         P: Into<ParParams>,
     {
@@ -430,20 +427,24 @@ where
             buf_size,
         } = params.into();
 
+        let (input_tx, input_rx) = utils::channel(buf_size);
         let (output_tx, output_rx) = utils::channel(buf_size);
-        let stream = self.shared();
+
+        rt::spawn(async move {
+            let _ = self.map(Ok).forward(input_tx.into_sink()).await;
+        });
 
         (0..num_workers).for_each(move |worker_index| {
             let output_tx = output_tx.clone();
             let f = f.clone();
-            let stream = stream.clone();
+            let input_rx = input_rx.clone();
 
             rt::spawn(async move {
                 let _ = stream::repeat(())
-                    .stateful_then((stream, f), |(stream, mut f), ()| async move {
-                        f(worker_index, stream)
+                    .stateful_then((input_rx, f), |(input_rx, mut f), ()| async move {
+                        f(worker_index, input_rx)
                             .await
-                            .map(move |(item, stream)| ((stream, f), item))
+                            .map(move |(item, input_rx)| ((input_rx, f), item))
                     })
                     .map(Ok)
                     .forward(output_tx.into_sink())
@@ -516,7 +517,7 @@ where
                 let fut = f(item);
                 Some((f, fut))
             })
-            .shared();
+            .spawned(buf_size);
 
         (0..num_workers).for_each(move |_| {
             let stream = stream.clone();
@@ -564,7 +565,7 @@ where
                 let func = f(item);
                 Some((f, func))
             })
-            .shared();
+            .spawned(buf_size);
         let (output_tx, output_rx) = utils::channel(buf_size);
 
         (0..num_workers).for_each(move |_| {
@@ -585,30 +586,31 @@ where
         output_rx.into_stream()
     }
 
-    fn par_reduce<N, F, Fut>(
+    fn par_reduce<P, F, Fut>(
         self,
-        num_workers: N,
+        params: P,
         reduce_fn: F,
     ) -> BoxFuture<'static, Option<Self::Item>>
     where
-        N: Into<NumWorkers>,
         F: 'static + FnMut(Self::Item, Self::Item) -> Fut + Send + Clone,
         Fut: 'static + Future<Output = Self::Item> + Send,
+        P: Into<ParParams>,
     {
-        let num_workers = num_workers.into().get();
-        let stream = self.shared();
+        let ParParams {
+            num_workers,
+            buf_size,
+        } = params.into();
+        let stream = self.spawned(buf_size);
 
         // phase 1
         let phase_1_future = {
             let reduce_fn = reduce_fn.clone();
             async move {
                 let reducer_futures = (0..num_workers).map(move |_| {
-                    let mut reduce_fn = reduce_fn.clone();
+                    let reduce_fn = reduce_fn.clone();
                     let stream = stream.clone();
 
-                    rt::spawn(
-                        async move { stream.reduce(|fold, item| reduce_fn(fold, item)).await },
-                    )
+                    rt::spawn(async move { stream.reduce(reduce_fn).await })
                 });
 
                 future::join_all(reducer_futures).await
@@ -624,11 +626,9 @@ where
 
             let mut count = 0;
 
-            for value in values {
-                if let Some(value) = value {
-                    feedback_tx.send_async(value).await.map_err(|_| ()).unwrap();
-                    count += 1;
-                }
+            for value in values.into_iter().flatten() {
+                feedback_tx.send_async(value).await.map_err(|_| ()).unwrap();
+                count += 1;
             }
 
             let pairing_future = rt::spawn(async move {
@@ -674,19 +674,22 @@ where
         phase_2_future.boxed()
     }
 
-    fn par_for_each<N, F, Fut>(self, num_workers: N, f: F) -> BoxFuture<'static, ()>
+    fn par_for_each<P, F, Fut>(self, params: P, f: F) -> BoxFuture<'static, ()>
     where
         F: 'static + FnMut(Self::Item) -> Fut + Send,
         Fut: 'static + Future<Output = ()> + Send,
-        N: Into<NumWorkers>,
+        P: Into<ParParams>,
     {
-        let num_workers = num_workers.into().get();
+        let ParParams {
+            num_workers,
+            buf_size,
+        } = params.into();
         let stream = self
             .stateful_map(f, |mut f, item| {
                 let fut = f(item);
                 Some((f, fut))
             })
-            .shared();
+            .spawned(buf_size);
 
         let worker_futures =
             (0..num_workers).map(move |_| rt::spawn(stream.clone().for_each(|fut| fut)));
@@ -694,19 +697,22 @@ where
         future::join_all(worker_futures).map(|_| ()).boxed()
     }
 
-    fn par_for_each_blocking<N, F, Func>(self, num_workers: N, f: F) -> BoxFuture<'static, ()>
+    fn par_for_each_blocking<P, F, Func>(self, params: P, f: F) -> BoxFuture<'static, ()>
     where
         F: 'static + FnMut(Self::Item) -> Func + Send,
         Func: 'static + FnOnce() + Send,
-        N: Into<NumWorkers>,
+        P: Into<ParParams>,
     {
-        let num_workers = num_workers.into().get();
+        let ParParams {
+            num_workers,
+            buf_size,
+        } = params.into();
         let stream = self
             .stateful_map(f, |mut f, item| {
                 let func = f(item);
                 Some((f, func))
             })
-            .shared();
+            .spawned(buf_size);
 
         let worker_futs: Vec<_> = (0..num_workers)
             .map(move |_| {
@@ -729,224 +735,226 @@ where
 #[cfg(test)]
 mod tests {
     use super::*;
+    use crate::utils::async_test;
     use rand::prelude::*;
     use std::time::Duration;
 
-    #[tokio::test]
-    async fn par_batching_test() {
-        let mut rng = rand::thread_rng();
-        let data: Vec<u32> = (0..10000).map(|_| rng.gen_range(0..10)).collect();
+    async_test! {
+        async fn par_batching_test() {
+            let mut rng = rand::thread_rng();
+            let data: Vec<u32> = (0..10000).map(|_| rng.gen_range(0..10)).collect();
 
-        let sums: Vec<_> = stream::iter(data)
-            .par_batching(None, |_, mut stream| async move {
-                let mut sum = stream.next().await?;
+            let sums: Vec<_> = stream::iter(data)
+                .par_batching(None, |_, rx| async move {
+                    let mut sum = rx.recv_async().await.ok()?;
 
-                while let Some(val) = stream.next().await {
-                    sum += val;
+                    while let Ok(val) = rx.recv_async().await {
+                        sum += val;
 
-                    if sum >= 1000 {
-                        return Some((sum, stream));
+                        if sum >= 1000 {
+                            return Some((sum, rx));
+                        }
                     }
-                }
 
-                None
-            })
-            .collect()
-            .await;
+                    None
+                })
+                .collect()
+                .await;
 
-        assert!(sums.iter().all(|&sum| sum >= 1000));
-    }
+            assert!(sums.iter().all(|&sum| sum >= 1000));
+        }
 
-    #[tokio::test]
-    async fn par_then_output_is_ordered_test() {
-        let max = 1000u64;
-        stream::iter(0..max)
-            .par_then(None, |value| async move {
-                rt::sleep(Duration::from_millis(value % 20)).await;
-                value
-            })
-            .fold(0u64, |expect, found| async move {
+
+        async fn par_then_output_is_ordered_test() {
+            let max = 1000u64;
+            stream::iter(0..max)
+                .par_then(None, |value| async move {
+                    rt::sleep(Duration::from_millis(value % 20)).await;
+                    value
+                })
+                .fold(0u64, |expect, found| async move {
+                    assert_eq!(expect, found);
+                    expect + 1
+                })
+                .await;
+        }
+
+
+        async fn par_then_unordered_test() {
+            let max = 1000u64;
+            let mut values: Vec<_> = stream::iter((0..max).into_iter())
+                .par_then_unordered(None, |value| async move {
+                    rt::sleep(Duration::from_millis(value % 20)).await;
+                    value
+                })
+                .collect()
+                .await;
+            values.sort();
+            values.into_iter().fold(0, |expect, found| {
                 assert_eq!(expect, found);
                 expect + 1
-            })
-            .await;
-    }
-
-    #[tokio::test]
-    async fn par_then_unordered_test() {
-        let max = 1000u64;
-        let mut values: Vec<_> = stream::iter((0..max).into_iter())
-            .par_then_unordered(None, |value| async move {
-                rt::sleep(Duration::from_millis(value % 20)).await;
-                value
-            })
-            .collect()
-            .await;
-        values.sort();
-        values.into_iter().fold(0, |expect, found| {
-            assert_eq!(expect, found);
-            expect + 1
-        });
-    }
-
-    #[tokio::test]
-    async fn par_reduce_test() {
-        {
-            let sum: Option<u64> = stream::iter(iter::empty())
-                .par_reduce(None, |lhs, rhs| async move { lhs + rhs })
-                .await;
-            assert!(sum.is_none());
+            });
         }
 
-        {
-            let max = 100_000u64;
-            let sum = stream::iter((1..=max).into_iter())
-                .par_reduce(None, |lhs, rhs| async move { lhs + rhs })
-                .await;
-            assert_eq!(sum, Some((1 + max) * max / 2));
+
+        async fn par_reduce_test() {
+            {
+                let sum: Option<u64> = stream::iter(iter::empty())
+                    .par_reduce(None, |lhs, rhs| async move { lhs + rhs })
+                    .await;
+                assert!(sum.is_none());
+            }
+
+            {
+                let max = 100_000u64;
+                let sum = stream::iter((1..=max).into_iter())
+                    .par_reduce(None, |lhs, rhs| async move { lhs + rhs })
+                    .await;
+                assert_eq!(sum, Some((1 + max) * max / 2));
+            }
         }
-    }
 
-    #[tokio::test]
-    async fn reorder_index_haling_test() {
-        let indexes = vec![5, 2, 1, 0, 6, 4, 3];
-        let output: Vec<_> = stream::iter(indexes)
-            .then(|index| async move {
-                rt::sleep(Duration::from_millis(20)).await;
-                (index, index)
-            })
-            .reorder_enumerated()
-            .collect()
-            .await;
-        assert_eq!(&output, &[0, 1, 2, 3, 4, 5, 6]);
-    }
 
-    #[tokio::test]
-    async fn enumerate_reorder_test() {
-        let max = 1000u64;
-        let iterator = (0..max).rev().step_by(2);
+        async fn reorder_index_haling_test() {
+            let indexes = vec![5, 2, 1, 0, 6, 4, 3];
+            let output: Vec<_> = stream::iter(indexes)
+                .then(|index| async move {
+                    rt::sleep(Duration::from_millis(20)).await;
+                    (index, index)
+                })
+                .reorder_enumerated()
+                .collect()
+                .await;
+            assert_eq!(&output, &[0, 1, 2, 3, 4, 5, 6]);
+        }
 
-        let lhs = stream::iter(iterator.clone())
-            .enumerate()
-            .par_then_unordered(None, |(index, value)| async move {
-                rt::sleep(std::time::Duration::from_millis(value % 20)).await;
-                (index, value)
-            })
-            .reorder_enumerated();
-        let rhs = stream::iter(iterator.clone());
 
-        let is_equal =
-            async_std::stream::StreamExt::all(&mut lhs.zip(rhs), |(lhs_value, rhs_value)| {
-                lhs_value == rhs_value
-            })
-            .await;
-        assert!(is_equal);
-    }
+        async fn enumerate_reorder_test() {
+            let max = 1000u64;
+            let iterator = (0..max).rev().step_by(2);
 
-    #[tokio::test]
-    async fn for_each_test() {
-        use std::sync::atomic::{self, AtomicUsize};
+            let lhs = stream::iter(iterator.clone())
+                .enumerate()
+                .par_then_unordered(None, |(index, value)| async move {
+                    rt::sleep(std::time::Duration::from_millis(value % 20)).await;
+                    (index, value)
+                })
+                .reorder_enumerated();
+            let rhs = stream::iter(iterator.clone());
 
-        {
-            let sum = Arc::new(AtomicUsize::new(0));
-            stream::iter(1..=1000)
-                .par_for_each(None, {
-                    let sum = sum.clone();
-                    move |value| {
-                        let sum = sum.clone();
-                        async move {
-                            sum.fetch_add(value, atomic::Ordering::SeqCst);
-                        }
-                    }
+            let is_equal =
+                async_std::stream::StreamExt::all(&mut lhs.zip(rhs), |(lhs_value, rhs_value)| {
+                    lhs_value == rhs_value
                 })
                 .await;
-            assert_eq!(sum.load(atomic::Ordering::SeqCst), (1 + 1000) * 1000 / 2);
+            assert!(is_equal);
         }
 
-        {
-            let sum = Arc::new(AtomicUsize::new(0));
-            stream::iter(1..=1000)
-                .par_for_each_blocking(None, {
-                    let sum = sum.clone();
-                    move |value| {
+
+        async fn for_each_test() {
+            use std::sync::atomic::{self, AtomicUsize};
+
+            {
+                let sum = Arc::new(AtomicUsize::new(0));
+                stream::iter(1..=1000)
+                    .par_for_each(None, {
                         let sum = sum.clone();
-                        move || {
-                            sum.fetch_add(value, atomic::Ordering::SeqCst);
+                        move |value| {
+                            let sum = sum.clone();
+                            async move {
+                                sum.fetch_add(value, atomic::Ordering::SeqCst);
+                            }
                         }
-                    }
-                })
-                .await;
-            assert_eq!(sum.load(atomic::Ordering::SeqCst), (1 + 1000) * 1000 / 2);
+                    })
+                    .await;
+                assert_eq!(sum.load(atomic::Ordering::SeqCst), (1 + 1000) * 1000 / 2);
+            }
+
+            {
+                let sum = Arc::new(AtomicUsize::new(0));
+                stream::iter(1..=1000)
+                    .par_for_each_blocking(None, {
+                        let sum = sum.clone();
+                        move |value| {
+                            let sum = sum.clone();
+                            move || {
+                                sum.fetch_add(value, atomic::Ordering::SeqCst);
+                            }
+                        }
+                    })
+                    .await;
+                assert_eq!(sum.load(atomic::Ordering::SeqCst), (1 + 1000) * 1000 / 2);
+            }
         }
-    }
 
-    #[tokio::test]
-    async fn tee_halt_test() {
-        let mut rx1 = stream::iter(0..).tee(1);
-        let mut rx2 = rx1.clone();
 
-        assert!(rx1.next().await.is_some());
-        assert!(rx2.next().await.is_some());
+        async fn tee_halt_test() {
+            let mut rx1 = stream::iter(0..).tee(1);
+            let mut rx2 = rx1.clone();
 
-        // drop rx1
-        drop(rx1);
+            assert!(rx1.next().await.is_some());
+            assert!(rx2.next().await.is_some());
 
-        // the following should not block
-        assert!(rx2.next().await.is_some());
-        assert!(rx2.next().await.is_some());
-        assert!(rx2.next().await.is_some());
-        assert!(rx2.next().await.is_some());
-        assert!(rx2.next().await.is_some());
-    }
+            // drop rx1
+            drop(rx1);
 
-    #[tokio::test]
-    async fn tee_test() {
-        let orig: Vec<_> = (0..100).collect();
+            // the following should not block
+            assert!(rx2.next().await.is_some());
+            assert!(rx2.next().await.is_some());
+            assert!(rx2.next().await.is_some());
+            assert!(rx2.next().await.is_some());
+            assert!(rx2.next().await.is_some());
+        }
 
-        let rx1 = stream::iter(orig.clone()).tee(1);
-        let rx2 = rx1.clone();
-        let rx3 = rx1.clone();
 
-        let fut1 = rx1
-            .then(|val| async move {
-                let millis = rand::thread_rng().gen_range(0..5);
-                rt::sleep(Duration::from_millis(millis)).await;
-                val
-            })
-            .collect();
-        let fut2 = rx2
-            .then(|val| async move {
-                let millis = rand::thread_rng().gen_range(0..5);
-                rt::sleep(Duration::from_millis(millis)).await;
-                val * 2
-            })
-            .collect();
-        let fut3 = rx3
-            .then(|val| async move {
-                let millis = rand::thread_rng().gen_range(0..5);
-                rt::sleep(Duration::from_millis(millis)).await;
-                val * 3
-            })
-            .collect();
+        async fn tee_test() {
+            let orig: Vec<_> = (0..100).collect();
 
-        let (vec1, vec2, vec3): (Vec<_>, Vec<_>, Vec<_>) = join!(fut1, fut2, fut3);
+            let rx1 = stream::iter(orig.clone()).tee(1);
+            let rx2 = rx1.clone();
+            let rx3 = rx1.clone();
 
-        // the collected method is possibly losing some of first few elements
-        let start1 = orig.len() - vec1.len();
-        let start2 = orig.len() - vec2.len();
-        let start3 = orig.len() - vec3.len();
+            let fut1 = rx1
+                .then(|val| async move {
+                    let millis = rand::thread_rng().gen_range(0..5);
+                    rt::sleep(Duration::from_millis(millis)).await;
+                    val
+                })
+                .collect();
+            let fut2 = rx2
+                .then(|val| async move {
+                    let millis = rand::thread_rng().gen_range(0..5);
+                    rt::sleep(Duration::from_millis(millis)).await;
+                    val * 2
+                })
+                .collect();
+            let fut3 = rx3
+                .then(|val| async move {
+                    let millis = rand::thread_rng().gen_range(0..5);
+                    rt::sleep(Duration::from_millis(millis)).await;
+                    val * 3
+                })
+                .collect();
 
-        assert!(orig[start1..]
-            .iter()
-            .zip(&vec1)
-            .all(|(&orig, &val)| orig == val));
-        assert!(orig[start2..]
-            .iter()
-            .zip(&vec2)
-            .all(|(&orig, &val)| orig * 2 == val));
-        assert!(orig[start3..]
-            .iter()
-            .zip(&vec3)
-            .all(|(&orig, &val)| orig * 3 == val));
+            let (vec1, vec2, vec3): (Vec<_>, Vec<_>, Vec<_>) = join!(fut1, fut2, fut3);
+
+            // the collected method is possibly losing some of first few elements
+            let start1 = orig.len() - vec1.len();
+            let start2 = orig.len() - vec2.len();
+            let start3 = orig.len() - vec3.len();
+
+            assert!(orig[start1..]
+                .iter()
+                .zip(&vec1)
+                .all(|(&orig, &val)| orig == val));
+            assert!(orig[start2..]
+                .iter()
+                .zip(&vec2)
+                .all(|(&orig, &val)| orig * 2 == val));
+            assert!(orig[start3..]
+                .iter()
+                .zip(&vec3)
+                .all(|(&orig, &val)| orig * 3 == val));
+        }
     }
 }
