@@ -1,9 +1,8 @@
 use crate::{
     common::*,
-    config::{BufSize, NumWorkers, ParParams},
+    config::{BufSize, ParParams},
     par_stream::ParStreamExt as _,
     rt,
-    shared_stream::Shared,
     stream::StreamExt as _,
     try_index_stream::TryIndexStreamExt as _,
     try_stream::{TakeUntilError, TryStreamExt as _},
@@ -36,10 +35,21 @@ where
     /// Fallible stream combinator for [par_batching](crate::ParStreamExt::par_batching).
     fn try_par_batching<U, P, F, Fut>(self, params: P, f: F) -> TryParBatching<U, Self::Error>
     where
-        F: 'static + Clone + Send + FnMut(usize, Shared<Self>) -> Fut,
-        Fut: 'static + Future<Output = Result<Option<(U, Shared<Self>)>, Self::Error>> + Send,
-        U: 'static + Send,
-        P: Into<ParParams>;
+        Self: Sized,
+        P: Into<ParParams>,
+        F: 'static
+            + Clone
+            + Send
+            + FnMut(usize, flume::Receiver<Result<Self::Ok, Self::Error>>) -> Fut,
+        Fut: 'static
+            + Future<
+                Output = Result<
+                    Option<(U, flume::Receiver<Result<Self::Ok, Self::Error>>)>,
+                    Self::Error,
+                >,
+            >
+            + Send,
+        U: 'static + Send;
 
     /// Fallible stream combinator for [par_then](crate::ParStreamExt::par_then).
     fn try_par_then<U, P, F, Fut>(
@@ -90,24 +100,24 @@ where
         Func: 'static + FnOnce() -> Result<U, Self::Error> + Send;
 
     /// Fallible stream combinator for [par_for_each](crate::par_stream::ParStreamExt::par_for_each).
-    fn try_par_for_each<N, F, Fut>(
+    fn try_par_for_each<P, F, Fut>(
         self,
-        num_workers: N,
+        params: P,
         f: F,
     ) -> BoxFuture<'static, Result<(), Self::Error>>
     where
-        N: Into<NumWorkers>,
+        P: Into<ParParams>,
         F: 'static + FnMut(Self::Ok) -> Fut + Send,
         Fut: 'static + Future<Output = Result<(), Self::Error>> + Send;
 
     /// Fallible stream combinator for [par_for_each_blocking](crate::par_stream::ParStreamExt::par_for_each_blocking).
-    fn try_par_for_each_blocking<N, F, Func>(
+    fn try_par_for_each_blocking<P, F, Func>(
         self,
-        num_workers: N,
+        params: P,
         f: F,
     ) -> BoxFuture<'static, Result<(), Self::Error>>
     where
-        N: Into<NumWorkers>,
+        P: Into<ParParams>,
         F: 'static + FnMut(Self::Ok) -> Func + Send,
         Func: 'static + FnOnce() -> Result<(), Self::Error> + Send;
 }
@@ -157,20 +167,34 @@ where
     where
         P: Into<ParParams>,
         U: 'static + Send,
-        F: 'static + Clone + Send + FnMut(usize, Shared<Self>) -> Fut,
-        Fut: 'static + Future<Output = Result<Option<(U, Shared<Self>)>, E>> + Send,
+        F: 'static
+            + Clone
+            + Send
+            + FnMut(usize, flume::Receiver<Result<Self::Ok, Self::Error>>) -> Fut,
+        Fut: 'static
+            + Future<
+                Output = Result<
+                    Option<(U, flume::Receiver<Result<Self::Ok, Self::Error>>)>,
+                    Self::Error,
+                >,
+            >
+            + Send,
     {
         let ParParams {
             num_workers,
             buf_size,
         } = params.into();
 
+        let (input_tx, input_rx) = utils::channel(buf_size);
         let (output_tx, output_rx) = utils::channel(buf_size);
         let (terminate_tx, _) = broadcast::channel(1);
-        let input_stream = self.shared();
+
+        rt::spawn(async move {
+            let _ = self.map(Ok).forward(input_tx.into_sink()).await;
+        });
 
         (0..num_workers).for_each(move |worker_index| {
-            let input_stream = input_stream.clone();
+            let input_rx = input_rx.clone();
             let output_tx = output_tx.clone();
             let mut terminate_rx = terminate_tx.subscribe();
             let terminate_tx = terminate_tx.clone();
@@ -182,18 +206,18 @@ where
                         let _ = terminate_rx.recv().await;
                     })
                     .stateful_then(
-                        Some((f, terminate_tx, input_stream)),
+                        Some((f, terminate_tx, input_rx)),
                         move |state, ()| async move {
-                            let (mut f, terminate_tx, input_stream) = state.unwrap();
-                            let result = f(worker_index, input_stream).await;
+                            let (mut f, terminate_tx, input_rx) = state.unwrap();
+                            let result = f(worker_index, input_rx).await;
 
                             if result.is_err() {
                                 let _ = terminate_tx.send(());
                             }
 
                             match result {
-                                Ok(Some((item, input_stream))) => {
-                                    Some((Some((f, terminate_tx, input_stream)), Ok(item)))
+                                Ok(Some((item, input_rx))) => {
+                                    Some((Some((f, terminate_tx, input_rx)), Ok(item)))
                                 }
                                 Ok(None) => None,
                                 Err(err) => Some((None, Err(err))),
@@ -302,13 +326,16 @@ where
         .boxed()
     }
 
-    fn try_par_for_each<N, F, Fut>(self, num_workers: N, f: F) -> BoxFuture<'static, Result<(), E>>
+    fn try_par_for_each<P, F, Fut>(self, params: P, f: F) -> BoxFuture<'static, Result<(), E>>
     where
-        N: Into<NumWorkers>,
+        P: Into<ParParams>,
         F: 'static + FnMut(T) -> Fut + Send,
         Fut: 'static + Future<Output = Result<(), E>> + Send,
     {
-        let num_workers = num_workers.into().get();
+        let ParParams {
+            num_workers,
+            buf_size,
+        } = params.into();
         let (terminate_tx, mut terminate_rx) = broadcast::channel(1);
         let input_stream = self
             .take_until_error()
@@ -319,7 +346,7 @@ where
                 let fut = item.map(|item| f(item));
                 Some((f, fut))
             })
-            .shared();
+            .spawned(buf_size);
 
         let worker_futures = (0..num_workers).map(move |_| {
             let terminate_tx = terminate_tx.clone();
@@ -349,17 +376,20 @@ where
             .boxed()
     }
 
-    fn try_par_for_each_blocking<N, F, Func>(
+    fn try_par_for_each_blocking<P, F, Func>(
         self,
-        num_workers: N,
+        params: P,
         f: F,
     ) -> BoxFuture<'static, Result<(), E>>
     where
-        N: Into<NumWorkers>,
+        P: Into<ParParams>,
         F: 'static + FnMut(T) -> Func + Send,
         Func: 'static + FnOnce() -> Result<(), E> + Send,
     {
-        let num_workers = num_workers.into().get();
+        let ParParams {
+            num_workers,
+            buf_size,
+        } = params.into();
         let (terminate_tx, mut terminate_rx) = broadcast::channel(1);
         let stream = self
             .take_until_error()
@@ -370,26 +400,21 @@ where
                 let fut = item.map(|item| f(item));
                 Some((f, fut))
             })
-            .shared();
+            .spawned(buf_size);
 
         let worker_futures = (0..num_workers).map(|_| {
             let mut stream = stream.clone();
             let terminate_tx = terminate_tx.clone();
 
             rt::spawn_blocking(move || {
-                loop {
-                    match rt::block_on(stream.next()) {
-                        Some(func) => {
-                            let result = (move || {
-                                (func?)()?;
-                                Ok(())
-                            })();
-                            if let Err(err) = result {
-                                let _result = terminate_tx.send(()); // shutdown workers
-                                return Err(err); // return error
-                            }
-                        }
-                        None => break,
+                while let Some(func) = rt::block_on(stream.next()) {
+                    let result = (move || {
+                        (func?)()?;
+                        Ok(())
+                    })();
+                    if let Err(err) = result {
+                        let _result = terminate_tx.send(()); // shutdown workers
+                        return Err(err); // return error
                     }
                 }
 
@@ -408,266 +433,268 @@ where
 #[cfg(test)]
 mod tests {
     use super::*;
+    use crate::utils::async_test;
     use rand::prelude::*;
 
-    #[tokio::test]
-    async fn try_par_batching_test() {
-        {
-            let mut stream = stream::iter(iter::repeat(1).take(10))
-                .map(Ok)
-                .try_par_batching(None, |_, _| async move {
-                    Result::<Option<((), _)>, _>::Err("init error")
-                });
+    async_test! {
+        async fn try_par_batching_test() {
+            {
+                let mut stream = stream::iter(iter::repeat(1).take(10))
+                    .map(Ok)
+                    .try_par_batching(None, |_, _| async move {
+                        Result::<Option<((), _)>, _>::Err("init error")
+                    });
 
-            assert_eq!(stream.next().await, Some(Err("init error")));
-            assert!(stream.next().await.is_none());
-        }
-
-        {
-            let mut stream = stream::repeat(1)
-                .take(10)
-                .map(Result::<_, ()>::Ok)
-                .try_par_batching(None, |_, mut stream| async move {
-                    let mut sum = 0;
-
-                    while let Some(val) = stream.next().await {
-                        sum += val?;
-                        if sum >= 3 {
-                            return Ok(Some((sum, stream)));
-                        }
-                    }
-
-                    if sum > 0 {
-                        return Ok(Some((sum, stream)));
-                    }
-
-                    Ok(None)
-                });
-
-            let mut total = 0;
-            while total < 10 {
-                let sum = stream.next().await.unwrap().unwrap();
-                assert!(sum <= 3);
-                total += sum;
+                assert_eq!(stream.next().await, Some(Err("init error")));
+                assert!(stream.next().await.is_none());
             }
-            assert!(stream.next().await.is_none());
-        }
 
-        {
-            let mut stream = stream::repeat(1).take(10).map(Ok).try_par_batching(
-                None,
-                |_, mut stream| async move {
-                    let mut sum = 0;
+            {
+                let mut stream = stream::repeat(1)
+                    .take(10)
+                    .map(Result::<_, ()>::Ok)
+                    .try_par_batching(None, |_, rx| async move {
+                        let mut sum = 0;
 
-                    while let Some(val) = stream.next().await {
-                        sum += val?;
-                        if sum >= 3 {
-                            return Ok(Some((sum, stream)));
+                        while let Ok(val) = rx.recv_async().await {
+                            sum += val?;
+                            if sum >= 3 {
+                                return Ok(Some((sum, rx)));
+                            }
+                        }
+
+                        if sum > 0 {
+                            return Ok(Some((sum, rx)));
                         }
-                    }
 
-                    if sum == 0 {
                         Ok(None)
-                    } else {
-                        Err(sum)
-                    }
-                },
-            );
+                    });
 
-            let mut total = 0;
-            while total < 10 {
-                let result = stream.next().await.unwrap();
-                match result {
-                    Ok(sum) => {
-                        assert!(sum == 3);
-                        total += sum;
-                    }
-                    Err(sum) => {
-                        assert!(sum < 3);
-                        break;
+                let mut total = 0;
+                while total < 10 {
+                    let sum = stream.next().await.unwrap().unwrap();
+                    assert!(sum <= 3);
+                    total += sum;
+                }
+                assert!(stream.next().await.is_none());
+            }
+
+            {
+                let mut stream = stream::repeat(1).take(10).map(Ok).try_par_batching(
+                    None,
+                    |_, rx| async move {
+                        let mut sum = 0;
+
+                        while let Ok(val) = rx.recv_async().await {
+                            sum += val?;
+                            if sum >= 3 {
+                                return Ok(Some((sum, rx)));
+                            }
+                        }
+
+                        if sum == 0 {
+                            Ok(None)
+                        } else {
+                            Err(sum)
+                        }
+                    },
+                );
+
+                let mut total = 0;
+                while total < 10 {
+                    let result = stream.next().await.unwrap();
+                    match result {
+                        Ok(sum) => {
+                            assert!(sum == 3);
+                            total += sum;
+                        }
+                        Err(sum) => {
+                            assert!(sum < 3);
+                            break;
+                        }
                     }
                 }
+                assert!(stream.next().await.is_none());
             }
-            assert!(stream.next().await.is_none());
-        }
-    }
-
-    #[tokio::test]
-    async fn try_par_for_each_test() {
-        {
-            let result = stream::iter(vec![Ok(1usize), Ok(2), Ok(6), Ok(4)].into_iter())
-                .try_par_for_each(None, |_| async move { Result::<_, ()>::Ok(()) })
-                .await;
-
-            assert_eq!(result, Ok(()));
         }
 
-        {
-            let result = stream::iter(vec![Ok(1usize), Ok(2), Err(-3isize), Ok(4)].into_iter())
-                .try_par_for_each(None, |_| async move { Ok(()) })
-                .await;
 
-            assert_eq!(result, Err(-3));
-        }
-    }
+        async fn try_par_for_each_test() {
+            {
+                let result = stream::iter(vec![Ok(1usize), Ok(2), Ok(6), Ok(4)].into_iter())
+                    .try_par_for_each(None, |_| async move { Result::<_, ()>::Ok(()) })
+                    .await;
 
-    #[tokio::test]
-    async fn try_par_for_each_blocking_test() {
-        {
-            let result = stream::iter(vec![Ok(1usize), Ok(2), Ok(6), Ok(4)])
-                .try_par_for_each_blocking(None, |_| || Result::<_, ()>::Ok(()))
-                .await;
+                assert_eq!(result, Ok(()));
+            }
 
-            assert_eq!(result, Ok(()));
+            {
+                let result = stream::iter(vec![Ok(1usize), Ok(2), Err(-3isize), Ok(4)].into_iter())
+                    .try_par_for_each(None, |_| async move { Ok(()) })
+                    .await;
+
+                assert_eq!(result, Err(-3));
+            }
         }
 
-        {
-            let result = stream::iter(0..)
-                .then(|val| async move {
-                    if val == 3 {
-                        Err(val)
-                    } else {
-                        Ok(val)
-                    }
-                })
-                .try_par_for_each_blocking(8, |_| || Ok(()))
-                .await;
 
-            assert_eq!(result, Err(3));
-        }
+        async fn try_par_for_each_blocking_test() {
+            {
+                let result = stream::iter(vec![Ok(1usize), Ok(2), Ok(6), Ok(4)])
+                    .try_par_for_each_blocking(None, |_| || Result::<_, ()>::Ok(()))
+                    .await;
 
-        {
-            let result = stream::iter(0..)
-                .map(Ok)
-                .try_par_for_each_blocking(None, |val| {
-                    move || {
+                assert_eq!(result, Ok(()));
+            }
+
+            {
+                let result = stream::iter(0..)
+                    .then(|val| async move {
                         if val == 3 {
-                            std::thread::sleep(Duration::from_millis(100));
                             Err(val)
                         } else {
-                            Ok(())
+                            Ok(val)
                         }
-                    }
-                })
-                .await;
+                    })
+                    .try_par_for_each_blocking(8, |_| || Ok(()))
+                    .await;
 
-            assert_eq!(result, Err(3));
+                assert_eq!(result, Err(3));
+            }
+
+            {
+                let result = stream::iter(0..)
+                    .map(Ok)
+                    .try_par_for_each_blocking(None, |val| {
+                        move || {
+                            if val == 3 {
+                                std::thread::sleep(Duration::from_millis(100));
+                                Err(val)
+                            } else {
+                                Ok(())
+                            }
+                        }
+                    })
+                    .await;
+
+                assert_eq!(result, Err(3));
+            }
         }
-    }
 
-    #[tokio::test]
-    async fn try_par_then_test() {
-        {
-            let vec: Vec<Result<_, _>> =
-                stream::iter(vec![Ok(1usize), Ok(2), Err(-3isize), Ok(4)].into_iter())
+
+        async fn try_par_then_test() {
+            {
+                let vec: Vec<Result<_, _>> =
+                    stream::iter(vec![Ok(1usize), Ok(2), Err(-3isize), Ok(4)].into_iter())
                     .try_par_then(None, |value| future::ok(value))
                     .collect()
                     .await;
 
-            assert!(matches!(
-                *vec,
-                [Err(-3)] | [Ok(1), Err(-3)] | [Ok(2), Err(-3)] | [Ok(1), Ok(2), Err(-3)],
-            ));
-        }
+                assert!(matches!(
+                    *vec,
+                    [Err(-3)] | [Ok(1), Err(-3)] | [Ok(2), Err(-3)] | [Ok(1), Ok(2), Err(-3)],
+                ));
+            }
 
-        {
-            let vec: Result<Vec<()>, ()> = stream::iter(vec![])
-                .try_par_then(None, |()| async move { Ok(()) })
-                .try_collect()
-                .await;
+            {
+                let vec: Result<Vec<()>, ()> = stream::iter(vec![])
+                    .try_par_then(None, |()| async move { Ok(()) })
+                    .try_collect()
+                    .await;
 
-            assert!(matches!(vec, Ok(vec) if vec.is_empty()));
-        }
+                assert!(matches!(vec, Ok(vec) if vec.is_empty()));
+            }
 
-        {
-            let vec: Vec<Result<_, _>> = stream::iter(1..)
-                .map(Ok)
-                .try_par_then(3, |index| async move {
-                    match index {
-                        3 | 6 => Err(index),
-                        index => Ok(index),
-                    }
-                })
-                .collect()
-                .await;
-
-            assert!(matches!(
-                *vec,
-                [Err(3)] | [Ok(1), Err(3)] | [Ok(2), Err(3)] | [Ok(1), Ok(2), Err(3)],
-            ));
-        }
-    }
-
-    #[tokio::test]
-    async fn try_reorder_enumerated_test() {
-        let len: usize = 1000;
-        let mut rng = rand::thread_rng();
-
-        for _ in 0..10 {
-            let err_index_1 = rng.gen_range(0..len);
-            let err_index_2 = rng.gen_range(0..len);
-            let min_err_index = err_index_1.min(err_index_2);
-
-            let results: Vec<_> = stream::iter(0..len)
-                .map(move |value| {
-                    if value == err_index_1 || value == err_index_2 {
-                        Err(-(value as isize))
-                    } else {
-                        Ok(value)
-                    }
-                })
-                .try_enumerate()
-                .try_par_then_unordered(None, |(index, value)| async move {
-                    rt::sleep(Duration::from_millis(value as u64 % 10)).await;
-                    Ok((index, value))
-                })
-                .try_reorder_enumerated()
-                .collect()
-                .await;
-            assert!(results.len() <= min_err_index + 1);
-
-            let (is_fused_at_error, _, _) = results.iter().cloned().fold(
-                (true, false, 0),
-                |(is_correct, found_err, expect), result| {
-                    if !is_correct {
-                        return (false, found_err, expect);
-                    }
-
-                    match result {
-                        Ok(value) => {
-                            let is_correct = value < min_err_index && value == expect && !found_err;
-                            (is_correct, found_err, expect + 1)
+            {
+                let vec: Vec<Result<_, _>> = stream::iter(1..)
+                    .map(Ok)
+                    .try_par_then(3, |index| async move {
+                        match index {
+                            3 | 6 => Err(index),
+                            index => Ok(index),
                         }
-                        Err(value) => {
-                            let is_correct = (-value) as usize == min_err_index && !found_err;
-                            let found_err = true;
-                            (is_correct, found_err, expect + 1)
+                    })
+                    .collect()
+                    .await;
+
+                assert!(matches!(
+                    *vec,
+                    [Err(3)] | [Ok(1), Err(3)] | [Ok(2), Err(3)] | [Ok(1), Ok(2), Err(3)],
+                ));
+            }
+        }
+
+
+        async fn try_reorder_enumerated_test() {
+            let len: usize = 1000;
+            let mut rng = rand::thread_rng();
+
+            for _ in 0..10 {
+                let err_index_1 = rng.gen_range(0..len);
+                let err_index_2 = rng.gen_range(0..len);
+                let min_err_index = err_index_1.min(err_index_2);
+
+                let results: Vec<_> = stream::iter(0..len)
+                    .map(move |value| {
+                        if value == err_index_1 || value == err_index_2 {
+                            Err(-(value as isize))
+                        } else {
+                            Ok(value)
                         }
-                    }
-                },
-            );
-            assert!(is_fused_at_error);
-        }
-    }
+                    })
+                    .try_enumerate()
+                    .try_par_then_unordered(None, |(index, value)| async move {
+                        rt::sleep(Duration::from_millis(value as u64 % 10)).await;
+                        Ok((index, value))
+                    })
+                    .try_reorder_enumerated()
+                    .collect()
+                    .await;
+                assert!(results.len() <= min_err_index + 1);
 
-    #[tokio::test]
-    async fn try_map_blocking_test() {
-        {
-            let vec: Vec<_> = stream::iter(vec![Ok(1u64), Ok(2), Err(-3i64), Ok(4)])
-                .try_map_blocking(None, |val| Ok(val.pow(10)))
-                .collect()
-                .await;
+                let (is_fused_at_error, _, _) = results.iter().cloned().fold(
+                    (true, false, 0),
+                    |(is_correct, found_err, expect), result| {
+                        if !is_correct {
+                            return (false, found_err, expect);
+                        }
 
-            assert_eq!(vec, [Ok(1), Ok(1024), Err(-3)]);
+                        match result {
+                            Ok(value) => {
+                                let is_correct = value < min_err_index && value == expect && !found_err;
+                                (is_correct, found_err, expect + 1)
+                            }
+                            Err(value) => {
+                                let is_correct = (-value) as usize == min_err_index && !found_err;
+                                let found_err = true;
+                                (is_correct, found_err, expect + 1)
+                            }
+                        }
+                    },
+                );
+                assert!(is_fused_at_error);
+            }
         }
 
-        {
-            let vec: Vec<_> = stream::iter(vec![Ok(1i64), Ok(2), Err(-3i64), Ok(4)])
-                .try_map_blocking(None, |val| if val >= 2 { Err(-val) } else { Ok(val) })
-                .collect()
-                .await;
 
-            assert_eq!(vec, [Ok(1), Err(-2)]);
+        async fn try_map_blocking_test() {
+            {
+                let vec: Vec<_> = stream::iter(vec![Ok(1u64), Ok(2), Err(-3i64), Ok(4)])
+                    .try_map_blocking(None, |val| Ok(val.pow(10)))
+                    .collect()
+                    .await;
+
+                assert_eq!(vec, [Ok(1), Ok(1024), Err(-3)]);
+            }
+
+            {
+                let vec: Vec<_> = stream::iter(vec![Ok(1i64), Ok(2), Err(-3i64), Ok(4)])
+                    .try_map_blocking(None, |val| if val >= 2 { Err(-val) } else { Ok(val) })
+                    .collect()
+                    .await;
+
+                assert_eq!(vec, [Ok(1), Err(-2)]);
+            }
         }
     }
 }
